(*******************************************************************************
 
  Project: Formal Verification of Secure Forwarding Protocols

  Author:  Tobias Klenze, ETH Zurich <tobias.klenze@inf.ethz.ch>
           Christoph Sprenger, ETH Zurich <sprenger@inf.ethz.ch>
  Version: CSF.1.0
  Isabelle Version: Isabelle2020

  Copyright (c) 2021 Tobias Klenze, Christoph Sprenger
  Licence: Mozilla Public License 2.0 (MPL)

*******************************************************************************)

section\<open>Parametrized dataplane protocol for directed protocols\<close>
text\<open>This is an instance of the @{text "Parametrized_Dataplane_2"} model, specifically for protocols
that authorize paths in an undirected fashion. We specialize the @{text "hf_valid_generic"} check
to a still parametrized, but more concrete @{text "hf_valid"} check. The rest of the parameters 
remain abstract until a later instantiation with a concrete protocols (see the instances directory).

While both the models for undirected and directed protocols import assumptions from the theory
@{text "Network_Assumptions"}, they differ in strength: the assumptions made by undirected protocols
are strictly weaker, since the entire forwarding path is authorized by each AS, and not only the 
future path from the perspective of each AS. 
In addition, the specific conditions that instances have to verify differs between the undirected
and the directed setting (compare the locales @{text "dataplane_3_undirected"} and 
@{text "dataplane_3_directed"}).

This explains the need to split up the verification of the attacker event into two theories. Despite
the differences that concrete protocols may exhibit, these two theories suffice to show the crux
of the refinement proof. The instances merely have to show a set of static conditions\<close>

theory Parametrized_Dataplane_3_directed
  imports
    "Parametrized_Dataplane_2" "Network_Assumptions"
begin

(******************************************************************************)
subsection \<open>Hop validation check, authorized segments, and path extraction.\<close>
(******************************************************************************)
text\<open>First we define a locale that requires a number of functions. We will later extend this
to a locale @{text "dataplane_3_directed"}, which makes assumptions on how these functions operate. 
We separate the assumptions in order to make use of some auxiliary definitions defined in this locale. \<close>
locale dataplane_3_directed_defs = network_assums_direct _ _ _ auth_seg0
  for auth_seg0 :: "('ainfo \<times> 'aahi ahi_scheme list) set" +
\<comment> \<open>@{text "hf_valid"} is the check that every hop performs on its own and neighboring hop fields as well as on
ainfo and uinfo. Note that this includes checking the validity of the info fields. 
Right now, we have a restriction in the model that this check can not depend on the previous hop 
field (see @{text "COND_hf_valid_no_prev"}).\<close>
  fixes hf_valid :: "'ainfo \<Rightarrow> msgterm
    \<Rightarrow> ('aahi, 'uhi) HF option 
    \<Rightarrow> ('aahi, 'uhi) HF
    \<Rightarrow> ('aahi, 'uhi) HF option \<Rightarrow> bool"
\<comment> \<open>We need checkInfo only for the empty segment (ainfo, []) since according to the definition any such
    ainfo will be contained in the intruder knowledge. With checkInfo we can restrict this.\<close>
  and checkInfo :: "'ainfo \<Rightarrow> bool"
\<comment> \<open>extr extracts from a given hop validation field (HVF hf) the entire authenticated future path that 
is embedded in the HVF.\<close>
  and extr :: "msgterm \<Rightarrow> 'aahi ahi_scheme list"
\<comment> \<open>@{text "extr_ainfo"} extracts the authenticated info field (ainfo) from a given hop validation field.\<close>
  and extr_ainfo :: "msgterm \<Rightarrow> 'ainfo"
\<comment> \<open>@{text "ik_auth_ainfo"} extracts what msgterms the intruder can learn from analyzing a given 
authenticated info field. Note that currently we do not have a similar function for the 
unauthenticated info field @{text "uinfo"}. Protocols should thus only use that field with terms
that the intruder can already synthesize (such as Numbers).\<close>
  and ik_auth_ainfo :: "'ainfo \<Rightarrow> msgterm"
\<comment> \<open>@{text "ik_hf"} extracts what msgterms the intruder can learn from analyzing a given hop field; for instance,
the hop validation field HVF hf and the segment identifier UHI hf.\<close>
  and ik_hf :: "('aahi, 'uhi) HF \<Rightarrow> msgterm set"
begin

abbreviation hf_valid_generic :: "'ainfo \<Rightarrow> msgterm
    \<Rightarrow> ('aahi, 'uhi) HF list
    \<Rightarrow> ('aahi, 'uhi) HF option 
    \<Rightarrow> ('aahi, 'uhi) HF
    \<Rightarrow> ('aahi, 'uhi) HF option \<Rightarrow> bool" where
  "hf_valid_generic ainfo uinfo pas pre hf nxt \<equiv> hf_valid ainfo uinfo pre hf nxt"

definition hfs_valid_prefix_generic ::
      "'ainfo \<Rightarrow> msgterm \<Rightarrow> ('aahi, 'uhi) HF list \<Rightarrow> ('aahi, 'uhi) HF option \<Rightarrow> ('aahi, 'uhi) HF list \<Rightarrow> 
      ('aahi, 'uhi) HF option \<Rightarrow> ('aahi, 'uhi) HF list"where
  "hfs_valid_prefix_generic ainfo uinfo pas pre fut nxt \<equiv> 
  TW.takeW (\<lambda> pre hf nxt . hf_valid ainfo uinfo pre hf nxt) pre fut nxt"

declare hfs_valid_prefix_generic_def[simp]

lemma prefix_hfs_valid_prefix_generic: 
  "prefix (hfs_valid_prefix_generic ainfo uinfo pas pre fut nxt) fut"
  by(auto intro: TW.takeW_prefix)

lemma cons_hfs_valid_prefix_generic: "hf_valid_generic ainfo uinfo pas (head pas) hf1 (head fut)
\<Longrightarrow> hfs_valid_prefix_generic ainfo uinfo pas (head pas) (hf1 # fut) None = 
    hf1 # (hfs_valid_prefix_generic ainfo uinfo (hf1#pas) (Some hf1) fut None)"
  apply(auto simp only: TW.takeW_split_tail[where x=hf1] hfs_valid_prefix_generic_def)
  apply auto
      apply (simp add: TW.takeW.simps(1))+
  using head_cons apply fastforce
  by (metis head.simps(1) head_cons)

sublocale dataplane_2_defs _ _ _ auth_seg0 hf_valid_generic hfs_valid_prefix_generic checkInfo extr extr_ainfo ik_auth_ainfo ik_hf
  apply unfold_locales
  using prefix_hfs_valid_prefix_generic cons_hfs_valid_prefix_generic by blast+

abbreviation hfs_valid where 
  "hfs_valid ainfo uinfo pre l nxt \<equiv> TW.holds (hf_valid ainfo uinfo) pre l nxt"

abbreviation hfs_valid_prefix where   
  "hfs_valid_prefix ainfo uinfo pre l nxt \<equiv> TW.takeW (hf_valid ainfo uinfo) pre l nxt"

abbreviation hfs_valid_None where
  "hfs_valid_None ainfo uinfo l \<equiv> hfs_valid ainfo uinfo None l None"

abbreviation hfs_valid_None_prefix where
  "hfs_valid_None_prefix ainfo uinfo l \<equiv> hfs_valid_prefix ainfo uinfo None l None"

end

print_locale dataplane_3_directed_defs
locale dataplane_3_directed_ik_defs = dataplane_3_directed_defs _ _ _ _ hf_valid checkInfo extr extr_ainfo ik_auth_ainfo ik_hf for 
        hf_valid :: "'ainfo \<Rightarrow> msgterm \<Rightarrow> ('aahi, 'uhi) HF option \<Rightarrow> ('aahi, 'uhi) HF \<Rightarrow> ('aahi, 'uhi) HF option \<Rightarrow> bool"
    and checkInfo :: "'ainfo \<Rightarrow> bool"
    and extr :: "msgterm \<Rightarrow> 'aahi ahi_scheme list"
    and extr_ainfo :: "msgterm \<Rightarrow> 'ainfo"
    and ik_auth_ainfo :: "'ainfo \<Rightarrow> msgterm"
    and ik_hf :: "('aahi, 'uhi) HF \<Rightarrow> msgterm set"
 +
\<comment> \<open>@{text "ik_add"} is Additional Intruder Knowledge, such as hop authenticators in EPIC L1.\<close>
fixes ik_add :: "msgterm set"
\<comment> \<open>@{text "ik_oracle"} is another type of additional Intruder Knowledge. We use it to model the attacker's
ability to brute-force individual hop validation fields and segment identifiers.\<close>
  and ik_oracle :: "msgterm set"
\<comment> \<open>As @{text "ik_oracle"} gives the attacker direct access to hop validation fields that could be used to break
the property, we have to either restrict the scope of the property, or restrict the attacker such 
that he cannot use the oracle-obtained hop validation fields in packets whose path origin matches
the path origin of the oracle query. We choose the latter approach and fix a predicate @{text "no_oracle"}
that tells us if the oracle has not been queried for a path origin (ainfo, uinfo combination).
This is a prophecy variable.\<close>
  and no_oracle :: "'ainfo \<Rightarrow> msgterm \<Rightarrow> bool"
begin

lemma auth_seg2_elem: "\<lbrakk>(ainfo, hfs) \<in> auth_seg2; hf \<in> set hfs\<rbrakk> 
  \<Longrightarrow> \<exists>pre nxt uinfo . hf_valid ainfo uinfo pre hf nxt \<and> checkInfo ainfo \<and> (ainfo, AHIS hfs) \<in> auth_seg0"
  by (auto simp add: auth_seg2_def TW.holds_takeW_is_identity dest!: TW.holds_set_list)

print_locale dataplane_2_ik_defs
sublocale dataplane_2_ik_defs _ _ _ _ hfs_valid_prefix_generic checkInfo extr extr_ainfo ik_auth_ainfo ik_hf hf_valid_generic ik_add ik_oracle no_oracle
  by unfold_locales
end

(******************************************************************************)
subsection \<open>Assumptions of the parametrized model\<close>
(******************************************************************************)
text\<open>We now list the assumptions of this parametrized model. \<close>

print_locale dataplane_3_directed_ik_defs
locale dataplane_3_directed = dataplane_3_directed_ik_defs _ _ _ _ hf_valid checkInfo extr extr_ainfo ik_auth_ainfo ik_hf ik_add ik_oracle no_oracle
  for hf_valid  :: "'ainfo \<Rightarrow> msgterm
    \<Rightarrow> ('aahi, 'uhi) HF option 
    \<Rightarrow> ('aahi, 'uhi) HF
    \<Rightarrow> ('aahi, 'uhi) HF option \<Rightarrow> bool"
    and checkInfo :: "'ainfo \<Rightarrow> bool"
    and extr :: "msgterm \<Rightarrow> 'aahi ahi_scheme list"
    and extr_ainfo :: "msgterm \<Rightarrow> 'ainfo"
    and ik_auth_ainfo :: "'ainfo \<Rightarrow> msgterm"
    and ik_hf :: "('aahi, 'uhi) HF \<Rightarrow> msgterm set"
    and ik_add :: "msgterm set"
    and ik_oracle :: "msgterm set"
    and no_oracle :: "'ainfo \<Rightarrow> msgterm \<Rightarrow> bool" +

\<comment> \<open>A valid validation field that is contained in ik corresponds to an authorized hop field.
   (The notable exceptions being oracle-obtained validation fields.)
   This relates the result of @{text "ik_hf"} to its argument. @{text "ik_hf"} has to produce a msgterm that is either
   unique for each given hop field x, or it is only produced by an 'equivalence class' of hop fields
   such that either all of the hop fields of the class are authorized, or none are.
   While the extr function (constrained by assumptions below) also binds the hop information to the
   validation field, it does so only for AHI and AInfo, but not for UHI.\<close>
    assumes COND_ik_hf:
    "\<lbrakk>hf_valid ainfo uinfo pre hf nxt; ik_hf hf \<subseteq> analz ik; ik_auth_ainfo ainfo \<in> analz ik;
      no_oracle ainfo uinfo\<rbrakk>
      \<Longrightarrow> \<exists>hfs . hf \<in> set hfs \<and> (ainfo, hfs) \<in> auth_seg2"
\<comment> \<open>A valid validation field that can be synthesized from the initial intruder knowledge is already
    contained in the initial intruder knowledge if it belongs to an honest AS. This can be combined
   with the previous assumption.\<close> 
    and COND_honest_hf_analz: 
    "\<lbrakk>ASID (AHI hf) \<notin> bad; hf_valid ainfo uinfo pre hf nxt; ik_hf hf \<subseteq> synth (analz ik);
      no_oracle ainfo uinfo\<rbrakk>
      \<Longrightarrow> ik_hf hf \<subseteq> analz ik"
\<comment> \<open>A valid info field that can be synthesized from the initial intruder knowledge is already
    contained in the initial intruder knowlege.\<close> 
    and COND_ainfo_analz: 
    "\<lbrakk>hf_valid ainfo uinfo pre hf nxt; ik_auth_ainfo ainfo \<in> synth (analz ik)\<rbrakk>
      \<Longrightarrow> ik_auth_ainfo ainfo \<in> analz ik"
\<comment> \<open>Extracting the path from the validation field of the first hop field of some path l
    returns an extension of the AHI-level path of the valid prefix of l.\<close> 
    and COND_path_prefix_extr:
    "prefix (AHIS (hfs_valid_prefix ainfo uinfo pre l nxt))
            (extr_from_hd l)"
\<comment> \<open>Extracting the path from the validation field of the first hop field of a completely valid path l
    returns a prefix of the AHI-level path of l. Together with @{thm "COND_path_prefix_extr"}, this implies
    that extr of a completely valid path l is exactly the same AHI-level path as l (see lemma below).\<close> 
    and COND_extr_prefix_path:
    "\<lbrakk>hfs_valid ainfo uinfo pre l nxt; nxt = None\<rbrakk> \<Longrightarrow> prefix (extr_from_hd l) (AHIS l)"
\<comment> \<open>The validation check does not depend on the prev hop field. For up-segments this is fine, but 
    this is an assumption we may eventually get rid off when we verify down-segments.\<close>
    and COND_hf_valid_no_prev:
    "hf_valid ainfo uinfo pre hf nxt \<longleftrightarrow> hf_valid ainfo uinfo pre' hf nxt"
\<comment> \<open>A valid hop field is only valid for one specific uinfo.\<close>
    and COND_hf_valid_uinfo:
    "\<lbrakk>hf_valid ainfo uinfo pre hf nxt; hf_valid ainfo' uinfo' pre' hf nxt'\<rbrakk> 
    \<Longrightarrow> uinfo' = uinfo"
begin

lemma holds_path_eq_extr:
    "\<lbrakk>hfs_valid ainfo uinfo pre l nxt; nxt = None\<rbrakk> \<Longrightarrow> extr_from_hd l = AHIS l"
  using COND_extr_prefix_path COND_path_prefix_extr 
  by (metis TW.holds_implies_takeW_is_identity prefix_order.eq_iff)

(******************************************************************************)
subsection \<open>Lemmas that are needed for the refinement proof\<close>
(******************************************************************************)

lemma honest_hf_analz_subsetI: 
    "\<lbrakk>ASID (AHI hf) \<notin> bad; hf_valid ainfo uinfo prev hf nxt; ik_hf hf \<subseteq> synth (analz ik);
      no_oracle ainfo uinfo; t \<in> ik_hf hf\<rbrakk>
      \<Longrightarrow> t \<in> analz ik"
  using COND_honest_hf_analz subsetI by blast

lemma extr_from_hd_eq: "(l \<noteq> [] \<and> l' \<noteq> [] \<and> hd l = hd l') \<or> (l = [] \<and> l' = []) \<Longrightarrow> extr_from_hd l = extr_from_hd l'" 
  apply (cases l)
   apply auto
  apply(cases l')
  by auto

lemma path_prefix_extr_l:
    "\<lbrakk>hd l = hd l'; l' \<noteq> []\<rbrakk> \<Longrightarrow> prefix (AHIS (hfs_valid_prefix ainfo uinfo pre l nxt))
            (extr_from_hd l')"
  using COND_path_prefix_extr extr_from_hd.elims list.sel(1) not_prefix_cases prefix_Cons prefix_Nil
  by metis

lemma path_prefix_extr_l':
    "\<lbrakk>hd l = hd l'; l' \<noteq> []; hf = hd l'\<rbrakk> \<Longrightarrow> prefix (AHIS (hfs_valid_prefix ainfo uinfo pre l nxt))
            (extr (HVF hf))"
  using COND_path_prefix_extr extr_from_hd.elims list.sel(1) not_prefix_cases prefix_Cons prefix_Nil
  by metis  (*takes a few sec*)


lemma pfrag_extr_auth:
assumes "hf \<in> set p" and "(ainfo, p) \<in> auth_seg2"
shows "pfragment ainfo (extr (HVF hf)) auth_seg0"
proof -
  obtain uinfo where p_verified: "hfs_valid_None ainfo uinfo p" 
      using assms(2) auth_seg2_def TW.holds_takeW_is_identity by fastforce
  have "\<exists>xs. suffix (hf # xs) p" by (simp add: Cons_suffix_set assms)
  then obtain xs where suf: "suffix (hf # xs) p" by (auto)
  then have "pfragment ainfo (hf # xs) auth_seg2" using assms(2)
    apply - (* ugh, this is "apply -" ugly... *)
    by (rule pfragment_suffix_self[where ?l="p"], simp_all)
  then have frag: "pfragment ainfo (AHIS (hf # xs)) auth_seg0"
    by (rule pfragment_auth_seg20)

  have "\<exists> pre . hfs_valid ainfo uinfo pre (hf # xs) None" using p_verified suf by (rule TW.holds_suffix)
  then have "pfragment ainfo (extr_from_hd (hf # xs)) auth_seg0"
    using holds_path_eq_extr[symmetric] frag by force (* takes a few seconds *)
  then show ?thesis by simp
qed

lemma X_in_ik_is_auth:
  assumes "ik_hf hf1 \<subseteq> analz ik" and "ik_auth_ainfo ainfo \<in> analz ik" and "no_oracle ainfo uinfo"
  shows "pfragment ainfo (AHIS (hfs_valid_prefix ainfo uinfo
                pre
                (hf1 # fut)
                nxt))
              auth_seg0"
proof -
  let ?pFu = "hf1 # fut"
  let ?takW = "(hfs_valid_prefix ainfo uinfo pre ?pFu nxt)"
  have "prefix (AHIS (hfs_valid_prefix ainfo uinfo pre ?takW (TW.extract (hf_valid ainfo uinfo) pre ?pFu nxt)))
               (extr_from_hd ?takW)"
    by(auto simp add: COND_path_prefix_extr simp del: AHIS_def)
  then have "prefix (AHIS ?takW) (extr_from_hd ?takW)"
    by(simp add: TW.takeW_takeW_extract)
  moreover from assms have "pfragment ainfo (extr_from_hd ?takW) auth_seg0"
    by (auto simp add: TW.takeW_split_tail dest!: COND_ik_hf intro: pfrag_extr_auth)
  ultimately show ?thesis
    by(auto intro: pfragment_prefix elim!: prefixE)
qed


subsubsection \<open>Fragment is extendable\<close>
(******************************************************************************)

text \<open>makes sure that: the segment is terminated, i.e. the leaf AS's HF has Eo = None\<close>
fun terminated2 :: "('aahi, 'uhi) HF list \<Rightarrow> bool" where
  "terminated2 (hf#xs) \<longleftrightarrow> DownIF (AHI hf) = None \<or> ASID (AHI hf) \<in> bad"
| "terminated2 [] = True" (* we allow this as a special case*)

lemma terminated20: "terminated (AHIS m) \<Longrightarrow> terminated2 m" by(induction m, auto)

lemma cons_snoc: "\<exists>y ys. x # xs = ys @ [y]"
  by (metis append_butlast_last_id rev.simps(2) rev_is_Nil_conv)

lemma terminated2_suffix:
  "\<lbrakk>terminated2 l; l = zs @ x # xs; DownIF (AHI x) \<noteq> None; ASID (AHI x) \<notin> bad\<rbrakk> \<Longrightarrow> \<exists>y ys. zs = ys @ [y]"
  by(cases "zs") 
    (fastforce intro: cons_snoc)+

lemma attacker_modify_cutoff: "\<lbrakk>(info, zs@hf#ys) \<in> auth_seg0; ASID hf = a;
      ASID hf' = a; UpIF hf' = UpIF hf; a \<in> bad; ys' = hf'#ys\<rbrakk> \<Longrightarrow> (info, ys') \<in> auth_seg0"
  by(auto simp add: ASM_modify dest: ASM_cutoff)

lemma auth_seg2_ik_hf[elim]: "\<lbrakk>x \<in> ik_hf hf; hf \<in> set hfs; (ainfo, hfs) \<in> auth_seg2 \<rbrakk> \<Longrightarrow> x \<in> analz ik"
  by(auto 3 4 simp add: ik_def)

text \<open>This lemma proves that an attacker-derivable segment that starts with an attacker hop field,
and has a next hop field which belongs to an honest AS, when restricted to its valid prefix, is
authorized. Essentially this is the case because the hop field of the honest AS already contains
an interface identifier DownIF that points to the attacker-controlled AS. Thus, there must have been
some attacker-owned hop field on the original authorized path. Given the assumptions we make in the
directed setting, the attacker can make take a suffix of an authorized path, such that his hop field
is first on the path, and he can change his own hop field if his hop field is the first on the path,
thus, that segment is also authorized.\<close>
lemma fragment_with_Eo_Some_extendable:
  assumes "ik_hf hf2 \<subseteq> synth (analz ik)"
  and "ik_auth_ainfo ainfo \<in> synth (analz ik)"
  and "ASID (AHI hf1) \<in> bad"
  and "ASID (AHI hf2) \<notin> bad" (* if A2 is bad, we could use directly COND_extension. This is covered below *)
  and "hf_valid ainfo uinfo pre hf1 (Some hf2)"
  and "no_oracle ainfo uinfo"
  shows
   "pfragment ainfo
      (ifs_valid_prefix pre'
        (AHIS (hfs_valid_prefix ainfo uinfo
          pre
          (hf1 # hf2 # fut)
          None))
      None)
     auth_seg0"
proof(cases)
  assume "hf_valid ainfo uinfo (Some hf1) hf2 (head fut) 
        \<and> if_valid (Some (AHI hf1)) (AHI hf2) (AHIo (head fut))"
  
  then have hf2true: "hf_valid ainfo uinfo (Some hf1) hf2 (head fut)"
                     "if_valid (Some (AHI hf1)) (AHI hf2) (AHIo (head fut))" by blast+
  then have "\<exists> hfs . hf2 \<in> set hfs \<and> (ainfo, hfs) \<in> auth_seg2" 
      using assms by(auto intro!: COND_ik_hf honest_hf_analz_subsetI COND_ainfo_analz)
  then obtain hfs uinfo' where hfs_def: 
    "hf2 \<in> set hfs" "(ainfo, hfs) \<in> auth_seg2" "hfs_valid_None ainfo uinfo' hfs"
    using COND_ik_hf by(auto simp add: auth_seg2_def TW.holds_takeW_is_identity)
  
  have termianted_hfs: "terminated2 hfs" 
    using hfs_def(2) by (auto simp add: auth_seg2_def ASM_terminated intro: terminated20)

  have "\<exists>pref hf1' ys . hfs = pref@[hf1']@(hf2#ys)"
    using hf2true(2) assms(4) hfs_def(1) terminated2_suffix
    by(fastforce dest: split_list intro: termianted_hfs)
  then obtain pref hf1' ys where hfs_unfold: "hfs = pref@[hf1']@(hf2#ys)" by fastforce 
  
  have hf2_valid:  "hf_valid ainfo uinfo' (Some hf1') hf2 (head ys)" (*only needed to show uinfo'_eq *)
     and hf1'true: "hf_valid ainfo uinfo' (tail pref) hf1' (Some hf2)"
    apply(cases ys)
    using hfs_def(3)
    by (auto simp add: hfs_def hfs_unfold TW.holds_unfold_prelnil tail_snoc TW.holds.simps(1) 
                elim!: TW.holds_unfold_prexnxt' intro: rev_exhaust[where ?xs=pref])
    
  have uinfo'_eq: "uinfo' = uinfo"
    using hf2_valid hf2true(1) by(intro COND_hf_valid_uinfo)
  
  
  have if_valid_hf2hf1': "if_valid (Some (AHI hf1')) (AHI hf2) (head (AHIS ys))"
    apply(cases ys)
    using assms(4) hfs_def(2) ASM_if_valid TW.holds_unfold_prexnxt' TW.holds_unfold_prelnil
    by(fastforce simp add: hfs_unfold auth_seg2_def)+

  have "pfragment ainfo (AHIS (hfs_valid_prefix ainfo uinfo
          None
          (hf1' # hf2 # fut)
          None))
        auth_seg0"
    apply(rule X_in_ik_is_auth)
    using hfs_def(1,2) assms(2,5,6) by(fastforce simp add: hfs_unfold ik_def intro!: COND_ainfo_analz)+
  
  then show ?thesis 
    apply-
    apply(rule strip_ifs_valid_prefix)
    apply(erule pfragment_self_eq_nil)
    apply(auto simp add: TW.takeW_split_tail[where ?x=hf1'])
    using assms(3-5) hf2true(2) if_valid_hf2hf1' hf1'true 
    apply(auto elim!: attacker_modify_cutoff[where ?hf'="AHI hf1"]
     simp add: TW.takeW_split_tail COND_hf_valid_no_prev[where pre="Some hf1", where pre'="Some hf1'"])
    using COND_hf_valid_no_prev uinfo'_eq by blast+
next
  assume hf2false: "\<not>(hf_valid ainfo uinfo (Some hf1) hf2 (head fut) 
                        \<and> if_valid (Some (AHI hf1)) (AHI hf2) (AHIo (head fut)))"
  then show ?thesis 
    apply(cases "hf_valid ainfo uinfo pre hf1 (Some hf2) \<and> if_valid pre' (AHI hf1) (Some (AHI hf2))")
      subgoal apply(cases "hf_valid ainfo uinfo (Some hf1) hf2 (head fut)")
        subgoal using assms(3) by(auto simp add: TW.takeW_split_tail intro: ASM_singleton) (*takes a few seconds*)
        subgoal apply(cases fut)
          using assms(3)
          by(auto simp add: TW.takeW_split_tail[where ?x=hf1] TW.takeW.simps
                  intro: ASM_singleton intro!: strip_ifs_valid_prefix)
      done
    by auto(auto simp add: TW.takeW_split_tail[where ?x=hf1] TW.takeW_split_tail[where ?x=hf2] 
                       TW.takeW.simps ASM_singleton assms(3) strip_ifs_valid_prefix)
qed

subsubsection \<open>A1 and A2 collude to make a wormhole\<close>
(******************************************************************************)

text \<open>We lift @{text "extend_pfragment0"} to DP2.\<close>
lemma extend_pfragment2:
  assumes "pfragment ainfo 
(ifs_valid_prefix (Some (AHI hf1))
(AHIS (hfs_valid_prefix ainfo uinfo
                      (Some hf1)
                      (hf2 # fut)
                      nxt))
             None)
                    auth_seg0"
  assumes "hf_valid ainfo uinfo pre hf1 (Some hf2)"
  assumes "ASID (AHI hf1) \<in> bad"
  assumes "ASID (AHI hf2) \<in> bad"
  shows "pfragment ainfo 
(ifs_valid_prefix pre'
(AHIS (hfs_valid_prefix ainfo uinfo
                    pre
                    (hf1 # hf2 # fut)
                    nxt))
             None)
                  auth_seg0"
  using assms
  apply(auto simp add: TW.takeW_split_tail[where ?P="hf_valid ainfo uinfo"])
  by(auto simp add: TW.takeW_split_tail[where ?P="if_valid"] TW.takeW.simps(1)
                      intro: ASM_singleton extend_pfragment0 strip_ifs_valid_prefix)

(*we prove this for all ainfo (but at most one of them will fit)*)
(*the case distinctions we make on whether the owner of the first hf is an adversary: this is the AS *AFTER*
the AS that sent the packet (the sender AS we know to be malicious as per the evt_dispatch_ext2 event). *)
text\<open>This is the central lemma that we need to prove to show the refinement between this model and 
dp1. It states: If an attacker can synthesize a segment from his knowledge, and does not use a path
origin that was used to query the oracle, then the valid prefix of the segment is authorized. Thus,
the attacker cannot create any valid but unauthorized segments.\<close>
lemma ik_seg_is_auth:
  assumes "\<And>hf . hf \<in> set hfs \<Longrightarrow> ik_hf hf \<subseteq> synth (analz ik)" and 
          "ik_auth_ainfo ainfo \<in> synth (analz ik)" and "nxt = None" and "no_oracle ainfo uinfo"
  shows "pfragment ainfo
            (ifs_valid_prefix prev'
              (AHIS (hfs_valid_prefix ainfo uinfo pre hfs nxt))
             None)
                auth_seg0"
using assms
proof(induction pre hfs nxt arbitrary: prev' rule: TW.takeW.induct[where ?Pa="hf_valid ainfo uinfo"])
  case (1 _ _) (* LENGTH: 0 *)
  then show ?case using append_Nil ASM_empty pfragment_def Nil_is_map_conv TW.takeW.simps(1)
    by (metis AHIS_def)
next
  case (2 pre hf nxt)
  then show ?case 
  proof(cases) (* LENGTH: 1, ADVERSARY *)
    assume "ASID (AHI hf) \<in> bad" (*bad*)
    then show ?thesis apply-
      by(intro strip_ifs_valid_prefix)
        (auto simp add: pfragment_def ASM_singleton TW.takeW_singleton intro!: exI[of _ "[]"])
  next
    assume "ASID (AHI hf) \<notin> bad" (*not bad*) (* LENGTH: 1, HONEST *)
    then show ?thesis using 2 assms 
      apply(intro strip_ifs_valid_prefix)
      by (auto simp add: "2.prems"(1) COND_ainfo_analz COND_honest_hf_analz X_in_ik_is_auth
               simp del: AHIS_def)
  qed
next
  case (3 pre hf nxt) (* LENGTH: 1, SPURIOUS *)
  then show ?case
    by (intro strip_ifs_valid_prefix, simp add: TW.takeW_singleton)
next
  case (4 pre hf1 hf2 xs nxt) (*this is the crux of the proof*)
  then show ?case 
  proof(cases)
    assume hf1bad: "ASID (AHI hf1) \<in> bad" (* LENGTH: N, ADVERSARY *)
    then show ?thesis 
    proof(cases)
      assume hf2bad: "ASID (AHI hf2) \<in> bad" (* LENGTH: N, ADVERSARY & ADVERSARY (wormhole assumption2) *)
      show ?thesis 
        apply(intro extend_pfragment2)
        apply(intro 4(2))
        using 4(1,3-5) \<open>no_oracle ainfo uinfo\<close> by(auto intro: hf1bad hf2bad)
    next
      assume "ASID (AHI hf2) \<notin> bad" (* LENGTH: N, ADVERSARY & HONEST (fragment is extendable) *)
      then show ?thesis
        using 4(1,3-6) hf1bad by(auto 3 4 intro!: fragment_with_Eo_Some_extendable 
                                        simp del: hfs_valid_prefix_generic_def AHIS_def)
    qed
  next
    assume "ASID (AHI hf1) \<notin> bad" (* LENGTH: N, HONEST *)
    then show ?thesis using 4(1,3-6)
      by(intro strip_ifs_valid_prefix)
        (auto intro!: X_in_ik_is_auth simp del: hfs_valid_prefix_generic_def AHIS_def
                dest: COND_honest_hf_analz COND_ainfo_analz)
  qed
next
  case 5 (* LENGTH: N, SPURIOUS *)
  then show ?case 
    by(intro strip_ifs_valid_prefix, simp add: TW.takeW_two_or_more)
qed

sublocale dataplane_2  _ _ _ _ hfs_valid_prefix_generic _ _ _ _ _ _ _ _ hf_valid_generic
  apply unfold_locales
  using ik_seg_is_auth by simp

end
end
