theory TakeWhile_upd imports Tools
begin

(******************************************************************************)
section \<open>takeW, holds and extract: Applying context-sensitive checks on list elements\<close>
(******************************************************************************)

text\<open>This theory defines three functions, takeW, holds and extract. It is embedded in a locale
       that takes infodicate P as an input that works on three arguments: info, x, and z. x is an
       element of a list, while info is the left neighbour on that list and z is the right neighbour.
       They are all of the same type 'a, except that info and z are of 'a option type, since
       neighbours don't always exist at the beginning and the end of lists. The functions takeW and
       holds work on an 'a list (with an additional info and z 'a option parameter).
       Both repeatedly apply P on elements xi in the list with their neighbours as context:\<close>
text_raw \<open>\begin{verbatim}
          holds info (x1#x2#...#xn#[]) z = 
              P info x1 x2 /\ P x1 x2 x3 /\ ... /\ P (xn-2) (xn-1) xn /\ P xn-1 xn z
          takeW info (x1#x2#...#xn#[]) z = the prefix of the list for which 'holds' holds.
\end{verbatim}\<close>
text\<open>  extract is a function that returns the last element of the list, or z if the list is empty.

       @{text "holds_takeW_extract"} is an interesting lemma that relates all three functions.

       In our applications, we usually invoke takeW and holds with the parameters None l None, where
       l is a list of elements which we want to check for P (using their neighboring elements as 
       context). takeW and holds thus mostly have the info and z parameters for their recursive 
       definition and induction schemes.\<close>

locale TWu =
  fixes P :: "('b \<Rightarrow> 'a \<Rightarrow> 'a option \<Rightarrow> bool)"
  fixes upd :: "('b \<Rightarrow> 'a \<Rightarrow> 'b)"
begin

subsection\<open>Definitions\<close>

text\<open>Apply @{text "upds"} on a sequence\<close>
abbreviation upds :: "'b \<Rightarrow> 'a list \<Rightarrow> 'b" where
  "upds \<equiv> foldl upd"

fun upd_opt :: "('b \<Rightarrow> 'a option \<Rightarrow> 'b)" where
  "upd_opt info (Some hf) = upd info hf"
| "upd_opt info None = info"

text\<open>holds returns true iff every element of a list, together with its context, satisfies P.\<close>
fun holds :: "'b \<Rightarrow> 'a list \<Rightarrow> 'a option \<Rightarrow> bool"
where
  "holds info (x # y # ys) nxt \<longleftrightarrow> P info x (Some y) \<and> holds (upd info y) (y # ys) nxt"
| "holds info [x] nxt \<longleftrightarrow> P info x nxt"
| "holds info [] nxt \<longleftrightarrow> True"

text\<open>holds returns the longest prefix of a list for every element, together with its context, satisfies P.\<close>
function takeW :: "'b \<Rightarrow> 'a list \<Rightarrow> 'a option \<Rightarrow> 'a list" where
  "takeW _ [] _ = []"
| "P info x xo \<Longrightarrow> takeW info [x] xo = [x]"
| "\<not> P info x xo \<Longrightarrow> takeW info [x] xo = []"
| "P info x (Some y) \<Longrightarrow> takeW info (x # y # xs) xo = x # takeW (upd info y) (y # xs) xo"
| "\<not> P info x (Some y) \<Longrightarrow> takeW info (x # y # xs) xo = []"
apply auto
  by (metis remdups_adj.cases)
termination
  by lexicographic_order

text\<open>extract returns the last element of a list, or nxt if the list is empty.\<close>
fun "extract" :: "'b \<Rightarrow> 'a list \<Rightarrow> 'a option \<Rightarrow> 'a option"
where
  "extract info (x # y # ys) nxt = (if P info x (Some y) then extract (upd info y) (y # ys) nxt else Some x)"
| "extract info [x] nxt = (if P info x nxt then nxt else (Some x))"
| "extract info [] nxt = nxt"


subsection\<open>Lemmas\<close>

text \<open>Lemmas packing singleton and at least two element cases into a single equation.\<close>

lemma takeW_singleton:
  "takeW info [x] xo = (if P info x xo then [x] else [])"
by (simp)

lemma takeW_two_or_more:
  "takeW info (x # y # zs) xo = (if P info x (Some y) then x # takeW (upd info y) (y # zs) xo else [])"
by (simp)


text \<open>Some lemmas for splitting the tail of the list argument.\<close>

text \<open>Splitting lemma formulated with if-then-else rather than case.\<close>

lemma takeW_split_tail:
  "takeW info (x # xs) nxt =
     (if xs = []
      then (if P info x nxt then [x] else [])
      else (if P info x (Some (hd xs)) then x # takeW (upd info (hd xs)) xs nxt else []))"
by (cases xs, auto)

lemma extract_split_tail:
  "extract info (x # xs) nxt =
     (case xs of
          [] \<Rightarrow> (if P info x nxt then nxt else (Some x))
        | (y # ys) \<Rightarrow> (if P info x (Some y) then extract (upd info y) (y # ys) nxt else Some x))"
by (cases xs, auto)

lemma holds_split_tail:
  "holds info (x # xs) nxt \<longleftrightarrow>
      (case xs of
          [] \<Rightarrow> P info x nxt
        | (y # ys) \<Rightarrow> P info x (Some y) \<and> holds (upd info y) (y # ys) nxt)"
by (cases xs, auto)

lemma holds_Cons_P:
  "holds info (x # xs) nxt \<Longrightarrow> \<exists>y . P info x y"
by (cases xs, auto)

lemma holds_Cons_holds:
  "holds info (x # xs) nxt \<Longrightarrow> holds (upd_opt info (head xs)) xs nxt"
by (cases xs, auto)

lemmas tail_splitting_lemmas =
  extract_split_tail holds_split_tail

text \<open>Interaction between @{term holds}, @{term takeWhile}, and @{term extract}.\<close>
declare if_split_asm [split]

lemma holds_takeW_extract: "holds info (takeW info xs nxt) (extract info xs nxt)"
apply(induction info xs nxt rule: takeW.induct)
apply auto
subgoal for info x y ys
  apply(cases "ys")
  apply(simp_all)
  done
done

text \<open>Interaction of @{term holds}, @{term takeWhile}, and @{term extract}
with @{term append}.\<close>

(*
lemma takeW_append:
  "takeW info (xs @ ys) nxt =
     (let y = case ys of [] \<Rightarrow> nxt | x # _ \<Rightarrow> Some x in
     (let new_info = case xs of [] \<Rightarrow> info | _ \<Rightarrow> (Some (last xs)) in
        if holds info xs y then xs @ takeW new_info ys nxt
                           else takeW infooder ha xs y))"
apply(induction info xs nxt rule: takeW.induct)
apply (simp_all add: Let_def split: list.split)
done

*)
lemma holds_append:
  "holds info (xs @ ys) nxt =
   (case ys of [] \<Rightarrow> holds info xs nxt | x # _ \<Rightarrow> 
     holds info xs (Some x) \<and> 
     (case xs of [] \<Rightarrow> holds info ys nxt 
                | _ \<Rightarrow> holds (upds info (tl xs@[x])) ys nxt))"
  by(induction info xs nxt rule: takeW.induct)
    (auto split: list.split)
(*
lemma holds_append:
  "holds info (xs @ ys) nxt =
   (let after_xs = case ys of [] \<Rightarrow> nxt | x # _ \<Rightarrow> Some x in
     holds info xs after_xs \<and> 
      (case xs of [] \<Rightarrow> holds info ys nxt 
                 | _ \<Rightarrow> 
        (let new_info = case ys of [] \<Rightarrow> upds info (tl xs) | x # _ \<Rightarrow> upds info ((tl xs)@[x]) in
          holds new_info ys nxt)))"
apply(induction info xs nxt rule: takeW.induct)
apply (auto simp add: Let_def split: list.split)
  done
*)




lemma upds_snoc: "upds uinfo (xs@[x]) = upd (upds uinfo xs) x" 
  by simp





(*
corollary holds_cutoff:
  "holds info (l1@l2) nxt \<Longrightarrow> \<exists> nxt' . holds info l1 nxt'"
by (meson holds_append)
lemma extract_append:
  "extract info (xs @ ys) nxt =
     (let y = case ys of [] \<Rightarrow> nxt | x # _ \<Rightarrow> Some x in
     (let new_info = case xs of [] \<Rightarrow> info | _ \<Rightarrow> (Some (last xs)) in
        if holds info xs y then extract new_info ys nxt else extract info xs y))"
apply(induction info xs nxt rule: takeW.induct)
apply (simp_all add: Let_def split: list.split)
done
*)

lemma takeW_prefix:
  "prefix (takeW info l nxt) l"
by (induction info l nxt rule: takeW.induct) auto

lemma takeW_set: "t \<in> set (TWu.takeW P upd info l nxt) \<Longrightarrow> t \<in> set l"
by(auto intro: takeW_prefix elim: set_prefix)

lemma holds_implies_takeW_is_identity:
  "holds info l nxt \<Longrightarrow> takeW info l nxt = l"
by (induction info l nxt rule: takeW.induct) auto

(*even stronger...*)
lemma holds_takeW_is_identity[simp]:
  "takeW info l nxt = l \<longleftrightarrow> holds info l nxt"
  by (induction info l nxt rule: takeW.induct) auto


lemma takeW_takeW_extract:
  "takeW info (takeW info l nxt) (extract info l nxt)
 = takeW info l nxt"
using holds_takeW_extract holds_implies_takeW_is_identity
  by blast

(*
text\<open>Show the equivalence of two takeW with different infos\<close>
lemma takeW_info_eqI:
"\<lbrakk>\<And>x . l = [x] \<Longrightarrow> P info x nxt \<longleftrightarrow> P info' x nxt;
 \<And>x1 x2 l' . l = x1#x2#l' \<Longrightarrow> P info x1 (Some x2) \<longleftrightarrow> P info' x1 (Some x2)\<rbrakk> \<Longrightarrow>
 takeW info l nxt = takeW info' l nxt"
  apply(cases l)
  subgoal by auto
  subgoal for a list
    by(cases list, auto simp add: takeW_singleton takeW_split_tail)
  done

lemma takeW_replace_info:
"\<lbrakk>P info x1 n; n = ifhead xs nxt\<rbrakk> \<Longrightarrow> prefix (takeW info' (x1#xs) nxt) (takeW info (x1#xs) nxt)"
  apply(cases xs)
  by(auto simp add: takeW_singleton takeW_split_tail)
*)

(******************************************************************************)
subsubsection \<open>Holds unfolding\<close>
(******************************************************************************)
text \<open>This section contains various lemmas that show how one can deduce P info' x' nxt' for some of
        info' x' nxt' out of a list l, for which we know that holds info l nxt is true.\<close>

lemma holds_set_list: "\<lbrakk>holds info l nxt; x \<in> set l\<rbrakk> \<Longrightarrow> \<exists> p y . P p x y"
  apply(induction info l nxt rule: TWu.takeW.induct[where ?Pa="P"]) by auto

lemma holds_unfold: "holds info l None \<Longrightarrow>
  l = [] \<or>
  (\<exists> x . l = [x] \<and> P info x None) \<or>
  (\<exists> x y ys . l = (x#y#ys) \<and> P info x (Some y) \<and> holds (upd info y) (y#ys) None)"
apply auto by (meson holds.elims(2))

(*
lemma holds_unfold_infoxnxt:
  "\<lbrakk>suffix (x0#x1#x2#xs) l; holds info l nxt\<rbrakk>
 \<Longrightarrow> P (upd info x0) x1 (Some x2)"
  by (auto simp add: suffix_def TWu.holds_append)

*)

lemma holds_unfold_prexnxt': (*weakened!*)
  "\<lbrakk>holds info l nxt; l = (zs@(x1#x2#xs)); zs \<noteq> []\<rbrakk>
 \<Longrightarrow> P (upds info ((tl zs)@[x1])) x1 (Some x2)"
  apply(cases zs) apply simp
  apply(simp only: TWu.holds_append)
  by auto
(*
lemma holds_unfold_xz:
  "\<lbrakk>suffix (x1#x2#xs) l; holds info l nxt\<rbrakk> \<Longrightarrow> \<exists> info'. P info' x1 (Some x2)"
  by (auto simp add: suffix_def TWu.holds_append)

lemma holds_unfold_infox:
  "\<lbrakk>suffix (x1#x2#xs) l; holds info l nxt\<rbrakk> \<Longrightarrow> \<exists> nxt'. P (Some x1) x2 nxt'"
by (auto simp add: suffix_def TWu.holds_append dest: holds_Cons_P)

*)
lemma holds_suffix:
  "\<lbrakk>holds info l nxt; suffix l' l\<rbrakk> \<Longrightarrow> \<exists> info'. holds info' l' nxt"
  apply(auto simp add: suffix_def)
  apply(simp only: TWu.holds_append)
  apply(cases l') apply auto 
  subgoal for zs a list
    apply(cases zs) by auto
  done

lemma holds_unfold_prelnil:
  "\<lbrakk>holds info l nxt; l = (zs@(x1#[])); zs \<noteq> []\<rbrakk>
 \<Longrightarrow> P (upds info ((tl zs)@[x1])) x1 nxt"
  apply(cases zs) apply simp
  apply(simp only: TWu.holds_append)
  by auto

end
end