(*******************************************************************************
 
  Project: Formal Verification of Secure Forwarding Protocols

  Author:  Tobias Klenze, ETH Zurich <tobias.klenze@inf.ethz.ch>
  Version: CSF.1.0
  Isabelle Version: Isabelle2020

  Copyright (c) 2021 Tobias Klenze
  Licence: Mozilla Public License 2.0 (MPL)

*******************************************************************************)

section \<open>EPIC Level 1 in the Basic Attacker Model\<close>
theory EPIC_L1_BA
  imports
    "../Parametrized_Dataplane_3_directed"
    "../infrastructure/Keys"
begin

locale epic_l1_defs = network_assums_direct _ _ _ auth_seg0 
  for auth_seg0 :: "(msgterm \<times> ahi list) set"
begin

(******************************************************************************)
subsection \<open>Hop validation check and extract functions\<close>
(******************************************************************************)

text\<open>The predicate @{term "hf_valid"} is given to the concrete parametrized model as a parameter.
It ensures the authenticity of the hop authenticator in the hop field. The predicate takes an authenticated
info field (in this model always a numeric value, hence the matching on Num ts), an unauthenticated
info field uinfo, the hop field to be validated and in some cases the next hop field.

We distinguish if there is a next hop field (this yields the two cases below). If there is not, then
the hop authenticator @{text "\<sigma>"} simply consists of a MAC over the authenticated info field and the local
routing information of the hop, using the key of the hop to which the hop field belongs. If on the
other hand, there is a subsequent hop field, then the uhi field of that hop field is also included
in the MAC computation.

The hop authenticator @{text "\<sigma>"} is used to compute both the hop validation field and the uhi field.
The first is computed as a MAC over the path origin (pair of absolute timestamp ts and the relative
timestamp given in uinfo), using the hop authenticator as a key to the MAC. The hop authenticator
is not secret, and any end host can use it to create a valid hvf. The uhi field, according to the 
protocol description, is @{text "\<sigma>"} shortened to a few bytes. We model this as applying the hash on @{text "\<sigma>"}.

The predicate @{term "hf_valid"} checks if the hop authenticator, hvf and uhi field are computed
correctly.\<close>
fun hf_valid :: "msgterm \<Rightarrow> msgterm
    \<Rightarrow> (unit, msgterm) HF option 
    \<Rightarrow> (unit, msgterm) HF
    \<Rightarrow> (unit, msgterm) HF option \<Rightarrow> bool" where 
  "hf_valid (Num ts) uinfo _ \<lparr>AHI = ahi, UHI = uhi, HVF = x\<rparr> (Some \<lparr>AHI = ahi2, UHI = uhi2, HVF = x2\<rparr>) \<longleftrightarrow> 
    (\<exists>\<sigma> upif downif. \<sigma> = Mac[macKey (ASID ahi)] (L [Num ts, upif, downif, uhi2]) \<and>
          ASIF (DownIF ahi) downif \<and> ASIF (UpIF ahi) upif \<and> uhi = Hash \<sigma> \<and> x = Mac[\<sigma>] \<langle>Num ts, uinfo\<rangle>)"
| "hf_valid (Num ts) uinfo _ \<lparr>AHI = ahi, UHI = uhi, HVF = x\<rparr> None \<longleftrightarrow> 
    (\<exists>\<sigma> upif downif. \<sigma> = Mac[macKey (ASID ahi)] (L [Num ts, upif, downif]) \<and>
          ASIF (DownIF ahi) downif \<and> ASIF (UpIF ahi) upif \<and> uhi = Hash \<sigma> \<and> x = Mac[\<sigma>] \<langle>Num ts, uinfo\<rangle>)"
| "hf_valid _ _ _ _ _ = False"

text\<open>We can extract the entire path from the uhi field, since it includes the hop authenticator, 
which includes the local forwarding information as well as, recursively, all upstream hop 
authenticators and their hop information.
However, the parametrized model defines the extract function to operate on the hop validation field,
not the uhi field. We therefore define a separate function that extracts the path from a hvf. 
We can do so, as both hvf and uhi contain the hop authenticator.
Internally, that function uses @{term "extrUhi"}.\<close>
fun extrUhi :: "msgterm \<Rightarrow> ahi list" where
  "extrUhi (Hash (Mac[macKey asid] (L [ts, upif, downif, uhi2])))
 = \<lparr>UpIF = ASO upif, DownIF = ASO downif, ASID = asid\<rparr> # extrUhi uhi2"
| "extrUhi (Hash (Mac[macKey asid] (L [ts, upif, downif])))
 = [\<lparr>UpIF = ASO upif, DownIF = ASO downif, ASID = asid\<rparr>]"
| "extrUhi _ = []"

text\<open>This function extracts from a hop validation field (HVF hf) the entire path.\<close>
fun extr :: "msgterm \<Rightarrow> ahi list" where
  "extr (Mac[\<sigma>] _) = extrUhi (Hash \<sigma>)" 
   | "extr _ = []"

text\<open>Extract the authenticated info field from a hop validation field.\<close>
fun extr_ainfo :: "msgterm \<Rightarrow> msgterm" where 
  "extr_ainfo (Mac[Mac[macKey asid] (L (Num ts # xs))] _) = Num ts"
| "extr_ainfo _ = \<epsilon>"

abbreviation ik_auth_ainfo :: "msgterm \<Rightarrow> msgterm" where
  "ik_auth_ainfo \<equiv> id"

text\<open>An authenticated info field is always a number (corresponding to a timestamp).\<close>
abbreviation checkInfo where 
  "checkInfo ainfo \<equiv> (\<exists>ts. ainfo = Num ts)"

text\<open>When observing a hop field, an attacker learns the HVF and the UHI. The AHI only contains 
public information that are not terms.\<close>
fun ik_hf :: "(unit, msgterm) HF \<Rightarrow> msgterm set" where 
  "ik_hf hf = {HVF hf, UHI hf}"

abbreviation no_oracle where "no_oracle \<equiv> (\<lambda> _ _. True)"

text\<open>We now define useful properties of the above definition.\<close>
lemma hf_valid_invert:
  "hf_valid tsn uinfo prev hf mo \<longleftrightarrow>
   ((\<exists>ahi ahi2 \<sigma> ts upif downif asid x upif2 downif2 asid2 uhi uhi2 x2.
     hf = \<lparr>AHI = ahi, UHI = uhi, HVF = x\<rparr> \<and>
     ASID ahi = asid \<and> ASIF (DownIF ahi) downif \<and> ASIF (UpIF ahi) upif \<and>
     mo = Some \<lparr>AHI = ahi2, UHI = uhi2, HVF = x2\<rparr> \<and>
     ASID ahi2 = asid2 \<and> ASIF (DownIF ahi2) downif2 \<and> ASIF (UpIF ahi2) upif2 \<and>
     \<sigma> = Mac[macKey asid] (L [tsn, upif, downif, uhi2]) \<and>
     tsn = Num ts \<and>
     uhi = Hash \<sigma> \<and>
     x = Mac[\<sigma>] \<langle>tsn, uinfo\<rangle>)
 \<or> (\<exists>ahi \<sigma> ts upif downif asid uhi x.
     hf = \<lparr>AHI = ahi, UHI = uhi, HVF = x\<rparr> \<and>
     ASID ahi = asid \<and> ASIF (DownIF ahi) downif \<and> ASIF (UpIF ahi) upif \<and>
     mo = None \<and>
     \<sigma> = Mac[macKey asid] (L [tsn, upif, downif]) \<and>
     tsn = Num ts \<and>
     uhi = Hash \<sigma> \<and>
     x = Mac[\<sigma>] \<langle>tsn, uinfo\<rangle>)
    )"
  apply(auto elim!: hf_valid.elims) using option.exhaust ASIF.simps by metis+

lemma hf_valid_checkInfo[dest]: "hf_valid ainfo uinfo prev hf z \<Longrightarrow> checkInfo ainfo"
  by(auto simp add: hf_valid_invert)

lemma info_hvf: 
  assumes "hf_valid ainfo uinfo prev m z" "HVF m = Mac[\<sigma>] \<langle>ainfo', uinfo'\<rangle> \<or> hf_valid ainfo' uinfo' prev' m z'" 
  shows "uinfo = uinfo'" "ainfo' = ainfo"
  using assms by(auto simp add: hf_valid_invert)


(******************************************************************************)
subsection\<open>Definitions and properties of the added intruder knowledge\<close>
(******************************************************************************)
text\<open>Here we define a sets which are added to the intruder knowledge: @{text "ik_add"}, which contains hop
authenticators.\<close>

sublocale dataplane_3_directed_defs _ _ _ auth_seg0 hf_valid checkInfo extr extr_ainfo ik_auth_ainfo ik_hf
  by unfold_locales

declare TW.holds_set_list[dest]
declare TW.holds_takeW_is_identity[simp]
declare parts_singleton[dest]

text\<open>This additional Intruder Knowledge allows us to model the attacker's access not only to the 
hop validation fields and segment identifiers of authorized segments (which are already given in 
@{text "ik_auth_hfs"}), but to the underlying hop authenticators that are used to create them.\<close>

definition ik_add :: "msgterm set" where
  "ik_add \<equiv> { \<sigma> | ainfo uinfo l hf \<sigma>.  
                 (ainfo, l) \<in> auth_seg2 \<and> hf \<in> set l \<and> HVF hf = Mac[\<sigma>] \<langle>ainfo, uinfo\<rangle> }"

lemma ik_addI:
  "\<lbrakk>(ainfo, l) \<in> auth_seg2; hf \<in> set l; HVF hf = Mac[\<sigma>] \<langle>ainfo, uinfo\<rangle>\<rbrakk> \<Longrightarrow> \<sigma> \<in> ik_add"
  by(auto simp add: ik_add_def)

lemma ik_add_form: "t \<in> ik_add \<Longrightarrow> \<exists> asid l . t = Mac[macKey asid] l"
  by(auto simp add: ik_add_def auth_seg2_def hf_valid_invert dest!: TW.holds_set_list)

lemma parts_ik_add[simp]: "parts ik_add = ik_add"
  by (auto intro!: parts_Hash dest: ik_add_form)

abbreviation ik_oracle :: "msgterm set" where "ik_oracle \<equiv> {}"

(******************************************************************************)
subsection\<open>Properties of the intruder knowledge, including @{text "ik_add"} and  @{text "ik_oracle"}\<close>
(******************************************************************************)
text\<open>We now instantiate the parametrized model's definition of the intruder knowledge, using the
definitions of @{text "ik_add"} and  @{text "ik_oracle"} from above. We then prove the properties 
that we need to instantiate the @{text "dataplane_3_directed"} locale.\<close>
sublocale
  dataplane_3_directed_ik_defs _ _ _ auth_seg0 hf_valid checkInfo extr extr_ainfo ik_auth_ainfo 
                  ik_hf ik_add ik_oracle no_oracle
  by unfold_locales

lemma ik_auth_hfs_form: "t \<in> parts ik_auth_hfs \<Longrightarrow> \<exists> t' . t = Hash t'"
  apply auto apply(drule parts_singleton)
  by(auto simp add: auth_seg2_def hf_valid_invert dest!: TW.holds_set_list)

declare ik_auth_hfs_def[simp del]

lemma parts_ik_auth_hfs[simp]: "parts ik_auth_hfs = ik_auth_hfs"
  by (auto intro!: parts_Hash ik_auth_hfs_form)

text\<open>This lemma allows us not only to expand the definition of @{term "ik_auth_hfs"}, but also 
to obtain useful properties, such as a term being a Hash, and it being part of a valid hop field.\<close>
lemma ik_auth_hfs_simp: 
  "t \<in> ik_auth_hfs \<longleftrightarrow> (\<exists>t' . t = Hash t') \<and> (\<exists>hf . (t = HVF hf \<or> t = UHI hf)
                    \<and> (\<exists>hfs. hf \<in> set hfs \<and> (\<exists>ainfo . (ainfo, hfs) \<in> auth_seg2
                    \<and> (\<exists> prev nxt uinfo. hf_valid ainfo uinfo prev hf nxt))))" (is "?lhs \<longleftrightarrow> ?rhs")
proof 
  assume asm: "?lhs" 
  then obtain ainfo hf hfs where 
    dfs: "hf \<in> set hfs" "(ainfo, hfs) \<in> auth_seg2" "t = HVF hf \<or> t = UHI hf"
    by(auto simp add: ik_auth_hfs_def)
  then obtain uinfo where "hfs_valid_None ainfo uinfo hfs"  "(ainfo, AHIS hfs) \<in> auth_seg0"
    by(auto simp add: auth_seg2_def)
  then show "?rhs" using asm dfs by(fast intro: ik_auth_hfs_form)
qed(auto simp add: ik_auth_hfs_def)

(******************************************************************************)
subsubsection \<open>Properties of Intruder Knowledge\<close>
(******************************************************************************)
lemma auth_ainfo[dest]: "\<lbrakk>(ainfo, hfs) \<in> auth_seg2\<rbrakk> \<Longrightarrow> \<exists> ts . ainfo = Num ts"
  by(auto simp add: auth_seg2_def)

lemma Num_ik[intro]: "Num ts \<in> ik"
  by(auto simp add: ik_def)
    (auto simp add: auth_seg2_def TW.holds.simps(3) intro!: exI[of _ "[]"]) 

text \<open>There are no ciphertexts (or signatures) in @{term "parts ik"}. Thus, @{term "analz ik"}
and @{term "parts ik"} are identical.\<close>
lemma analz_parts_ik[simp]: "analz ik = parts ik"
  apply(rule no_crypt_analz_is_parts)
  by(auto simp add: ik_def auth_seg2_def ik_auth_hfs_simp)
    (auto 3 4 simp add: ik_add_def auth_seg2_def hf_valid_invert)

lemma parts_ik[simp]: "parts ik = ik"
  by(auto 3 4 simp add: ik_def auth_seg2_def)

lemma key_ik_bad: "Key (macK asid) \<in> ik \<Longrightarrow> asid \<in> bad"
  by(auto simp add: ik_def hf_valid_invert)
    (auto 3 4 simp add: auth_seg2_def ik_auth_hfs_simp ik_add_def hf_valid_invert)

(******************************************************************************)
subsubsection\<open>Updating hop fields with different uinfo\<close>
(******************************************************************************)
text\<open>Those hop validation fields contained in @{term "auth_seg2"} or that can be generated from the hop
authenticators in @{text "ik_add"} have the property that they are agnostic about the uinfo field. If a
hop validation field is contained in @{term "auth_seg2"} (resp. derivable from @{text "ik_add"}), 
then a field with a different uinfo is also contained (resp. derivable).
To show this, we first define a function that updates uinfo in a hop validation field.\<close>
fun uinfo_upd_hf :: "msgterm \<Rightarrow> (unit, msgterm) HF \<Rightarrow> (unit, msgterm) HF" where
  "uinfo_upd_hf new_uinfo hf = 
    (case HVF hf of Mac[\<sigma>] \<langle>ainfo, uinfo\<rangle> \<Rightarrow> hf\<lparr>HVF := Mac[\<sigma>] \<langle>ainfo, new_uinfo\<rangle>\<rparr> | _ \<Rightarrow> hf)"

fun uinfo_upd :: "msgterm \<Rightarrow> (unit, msgterm) HF list \<Rightarrow> (unit, msgterm) HF list" where 
  "uinfo_upd new_uinfo hfs = map (uinfo_upd_hf new_uinfo) hfs"

lemma uinfo_upd_valid: 
  "hfs_valid ainfo uinfo pre l nxt \<Longrightarrow> hfs_valid ainfo new_uinfo pre (uinfo_upd new_uinfo l) nxt"
  apply(induction pre l nxt rule: TW.holds.induct)
  apply auto
  subgoal for prev x y ys z
    by(cases "map (uinfo_upd_hf new_uinfo) ys")
      (auto simp add: TW.holds_split_tail hf_valid_invert)
  by(auto 3 4 simp add: TW.holds_split_tail hf_valid_invert TW.holds.simps(3))

lemma uinfo_upd_hf_AHI: "AHI (uinfo_upd_hf new_uinfo hf) = AHI hf"
  apply(cases "HVF hf") apply auto
  subgoal for x apply(cases x) apply auto
    subgoal for x1 x2 apply(cases x2) by auto
    done
  done

lemma uinfo_upd_hf_AHIS[simp]: "AHIS (map (uinfo_upd_hf new_uinfo) l) = AHIS l"
  apply(induction l) using uinfo_upd_hf_AHI by auto

lemma uinfo_upd_auth_seg2:
  assumes "hf_valid ainfo uinfo prev m z" "\<sigma> = Mac[Key (macK asid)] j"
          "HVF m = Mac[\<sigma>] \<langle>ainfo, uinfo'\<rangle>" "\<sigma> \<in> ik_add"
  shows "\<exists>hfs. m \<in> set hfs \<and> (ainfo, hfs) \<in> auth_seg2"
proof-
  from assms(4) obtain ainfo_add uinfo_add l_add hf_add where
    "(ainfo_add, l_add) \<in> auth_seg2" "hf_add \<in> set l_add" "HVF hf_add = Mac[\<sigma>] \<langle>ainfo_add, uinfo_add\<rangle>"
    by(auto simp add: ik_add_def)
  then have add: "m \<in> set (uinfo_upd uinfo l_add)" "(ainfo_add, (uinfo_upd uinfo l_add)) \<in> auth_seg2"
    using assms(1-3) apply(auto simp add: auth_seg2_def simp del: AHIS_def)
    apply(auto simp add: hf_valid_invert intro!: image_eqI dest!: TW.holds_set_list)[1]
    by(auto intro!: exI elim: ahi_eq dest: uinfo_upd_valid simp del: AHIS_def)
  then have "ainfo_add = ainfo" 
    using assms(1) by(auto simp add: auth_seg2_def dest!: TW.holds_set_list dest: info_hvf)
  then show ?thesis using add by fastforce
qed

lemma MAC_synth_helper:
"\<lbrakk>hf_valid ainfo uinfo prev m z; 
  HVF m = Mac[\<sigma>] \<langle>ainfo, uinfo\<rangle>; \<sigma> = Mac[Key (macK asid)] j; \<sigma> \<in> ik \<or> HVF m \<in> ik\<rbrakk>
       \<Longrightarrow> \<exists>hfs. m \<in> set hfs \<and> (ainfo, hfs) \<in> auth_seg2"
  apply(auto simp add: ik_def ik_auth_hfs_simp dest: ik_add_form)
  prefer 3 subgoal by(auto elim!: uinfo_upd_auth_seg2)
  prefer 3 subgoal by(auto elim!: uinfo_upd_auth_seg2 intro: ik_addI dest: info_hvf HOL.sym)
  by(auto simp add: hf_valid_invert)

text\<open>This definition helps with the limiting the number of cases generated. We don't require it, 
but it is convenient. Given a hop validation field and an asid, return if the hvf has the expected
format.\<close>
definition mac_format :: "msgterm \<Rightarrow> as \<Rightarrow> bool" where 
  "mac_format m asid \<equiv> \<exists> j ts uinfo . m = Mac[Mac[macKey asid] j] \<langle>Num ts, uinfo\<rangle>"

text\<open>If a valid hop field is derivable by the attacker, but does not belong to the attacker, then 
the hop field is already contained in the set of authorized segments.\<close>
lemma MAC_synth:
  assumes "hf_valid ainfo uinfo prev m z" "HVF m \<in> synth ik" "mac_format (HVF m) asid"
    "asid \<notin> bad" "checkInfo ainfo"
  shows "\<exists>hfs . m \<in> set hfs \<and> (ainfo, hfs) \<in> auth_seg2"
  using assms
  apply(auto simp add: mac_format_def elim!: MAC_synth_helper dest!: key_ik_bad)
  apply(auto simp add: ik_def ik_auth_hfs_simp dest: ik_add_form)
  using assms(1) by(auto dest: info_hvf simp add: hf_valid_invert)

(******************************************************************************)
subsection\<open>Direct proof goals for interpretation of @{text "dataplane_3_directed"}\<close>
(******************************************************************************)

lemma COND_honest_hf_analz:
  assumes "ASID (AHI hf) \<notin> bad" "hf_valid ainfo uinfo prev hf nxt" "ik_hf hf \<subseteq> synth (analz ik)"
    "no_oracle ainfo uinfo"
    shows "ik_hf hf \<subseteq> analz ik"
proof-
  let ?asid = "ASID (AHI hf)"
  from assms(3) have hf_synth_ik: "HVF hf \<in> synth ik" "UHI hf \<in> synth ik" by auto
  from assms(2) have "mac_format (HVF hf) ?asid" "checkInfo ainfo" 
    by(auto simp add: mac_format_def hf_valid_invert)
  then obtain hfs where "hf \<in> set hfs" "(ainfo, hfs) \<in> auth_seg2"
    using assms(1,2,4) hf_synth_ik by(auto dest!: MAC_synth)
  then have "HVF hf \<in> ik" "UHI hf \<in> ik" 
    using assms(2)
    by(auto simp add: ik_auth_hfs_def intro!: ik_ik_auth_hfs intro!: exI) 
  then show ?thesis by auto
qed

lemma COND_ainfo_analz:
  assumes "hf_valid ainfo uinfo prev hf nxt" and "ik_auth_ainfo ainfo \<in> synth (analz ik)"
  shows "ik_auth_ainfo ainfo \<in> analz ik"
  using assms by(auto simp add: hf_valid_invert)

lemma COND_ik_hf: 
  assumes "hf_valid ainfo uinfo prev hf z" and "HVF hf \<in> ik" and "no_oracle ainfo uinfo"
  shows "\<exists>hfs. hf \<in> set hfs \<and> (ainfo, hfs) \<in> auth_seg2"
proof-
  from assms have "checkInfo ainfo" by auto
  then obtain hfs ainfo where hfs_def: "hf \<in> set hfs" "(ainfo, hfs) \<in> auth_seg2"
  using assms by(auto 3 4 simp add: hf_valid_invert ik_auth_hfs_simp ik_def dest: ahi_eq
                             dest!: ik_add_form)
  then obtain hfs ainfo where hfs_def: "hf \<in> set hfs" "(ainfo, hfs) \<in> auth_seg2" by auto
  show ?thesis 
    using hfs_def apply (auto simp add: auth_seg2_def dest!: TW.holds_set_list)
    using hfs_def assms(1) by (auto simp add: auth_seg2_def dest: info_hvf)
  qed

lemma COND_extr_prefix_path:
  "\<lbrakk>hfs_valid ainfo uinfo pre l nxt; nxt = None\<rbrakk> \<Longrightarrow> prefix (extr_from_hd l) (AHIS l)"
  by(induction pre l nxt rule: TW.holds.induct)
    (auto simp add: TW.holds_split_tail TW.holds.simps(1) hf_valid_invert,
     auto split: list.split_asm simp add: hf_valid_invert intro!: ahi_eq elim: ASIF.elims)

lemma COND_path_prefix_extr:
  "prefix (AHIS (hfs_valid_prefix ainfo uinfo pre l nxt))
          (extr_from_hd l)"
  apply(induction pre l nxt rule: TW.takeW.induct[where ?Pa="hf_valid ainfo uinfo"])
  apply(auto simp add: TW.takeW_split_tail TW.takeW.simps(1))
  apply(auto simp add: hf_valid_invert intro!: ahi_eq) (*takes a few seconds*)
    by(auto elim: ASIF.elims)

lemma COND_hf_valid_no_prev:
  "hf_valid ainfo uinfo prev hf z \<longleftrightarrow> hf_valid ainfo uinfo prev' hf z"
  by(auto simp add: hf_valid_invert)

lemma COND_hf_valid_uinfo:
  "\<lbrakk>hf_valid ainfo uinfo pre hf nxt; hf_valid ainfo' uinfo' pre' hf nxt'\<rbrakk> \<Longrightarrow> uinfo' = uinfo"
  by(auto dest: info_hvf)

(******************************************************************************)
subsection\<open>Instantiation of @{text "dataplane_3_directed"} locale\<close>
(******************************************************************************)
sublocale
  dataplane_3_directed _ _ _ auth_seg0 hf_valid checkInfo extr extr_ainfo ik_auth_ainfo ik_hf ik_add 
            ik_oracle no_oracle
  apply unfold_locales
  using COND_ik_hf COND_honest_hf_analz COND_ainfo_analz COND_extr_prefix_path
  COND_path_prefix_extr COND_hf_valid_no_prev COND_hf_valid_uinfo by auto


end
end
