(*******************************************************************************
 
  Project: Formal Verification of Secure Forwarding Protocols

  Author:  Tobias Klenze, ETH Zurich <tobias.klenze@inf.ethz.ch>
  Version: CSF.1.0
  Isabelle Version: Isabelle2020

  Copyright (c) 2021 Tobias Klenze
  Licence: Mozilla Public License 2.0 (MPL)

*******************************************************************************)

chapter\<open>Instances\<close>
text\<open>Here we instantiate our concrete parametrized models with a number of protocols from the 
literature and variants of them that we derive ourselves.\<close>
section \<open>SCION\<close>
theory SCION
  imports
    "../Parametrized_Dataplane_3_directed"
    "../infrastructure/Keys"
begin

locale scion_defs = network_assums_direct _ _ _ auth_seg0
  for auth_seg0 :: "(msgterm \<times> ahi list) set"
begin

(******************************************************************************)
subsection \<open>Hop validation check and extract functions\<close>
(******************************************************************************)
type_synonym SCION_HF = "(unit, unit) HF"

text\<open>The predicate @{term "hf_valid"} is given to the concrete parametrized model as a parameter.
It ensures the authenticity of the hop authenticator in the hop field. The predicate takes an authenticated
info field (in this model always a numeric value, hence the matching on Num ts), the hop field to 
be validated and in some cases the next hop field.

We distinguish if there is a next hop field (this yields the two cases below). If there is not, then
the hvf simply consists of a MAC over the authenticated info field and the local
routing information of the hop, using the key of the hop to which the hop field belongs. If on the
other hand, there is a subsequent hop field, then the hvf of that hop field is also included
in the MAC computation.\<close>
fun hf_valid :: "msgterm \<Rightarrow> msgterm
    \<Rightarrow> SCION_HF option 
    \<Rightarrow> SCION_HF
    \<Rightarrow> SCION_HF option \<Rightarrow> bool" where 
  "hf_valid (Num ts) uinfo _ \<lparr>AHI = ahi, UHI = _, HVF = x\<rparr> (Some \<lparr>AHI = ahi2, UHI = _, HVF = x2\<rparr>) \<longleftrightarrow> 
    (\<exists>upif downif upif2 downif2. 
          x = Mac[macKey (ASID ahi)] (L [Num ts, upif, downif, upif2, downif2, x2]) \<and>
          ASIF (DownIF ahi) downif \<and> ASIF (UpIF ahi) upif \<and> 
          ASIF (DownIF ahi2) downif2 \<and> ASIF (UpIF ahi2) upif2 \<and> uinfo = \<epsilon>)"
| "hf_valid (Num ts) uinfo _ \<lparr>AHI = ahi, UHI = _, HVF = x\<rparr> None \<longleftrightarrow> 
    (\<exists>upif downif. x = Mac[macKey (ASID ahi)] (L [Num ts, upif, downif]) \<and>
          ASIF (DownIF ahi) downif \<and> ASIF (UpIF ahi) upif \<and> uinfo = \<epsilon>)"
| "hf_valid _ _ _ _ _ = False"

text\<open>We can extract the entire path from the hvf field, which includes the local forwarding of the
current hop, the local forwarding information of the next hop (if existant) and, recursively, all 
upstream hvf fields and their hop information.\<close>
fun extr :: "msgterm \<Rightarrow> ahi list" where
  "extr (Mac[macKey asid] (L [ts, upif, downif, upif2, downif2, x2]))
 = \<lparr>UpIF = ASO upif, DownIF = ASO downif, ASID = asid\<rparr> # extr x2"
| "extr (Mac[macKey asid] (L [ts, upif, downif]))
 = [\<lparr>UpIF = ASO upif, DownIF = ASO downif, ASID = asid\<rparr>]"
| "extr _ = []"

text\<open>Extract the authenticated info field from a hop validation field.\<close>
fun extr_ainfo :: "msgterm \<Rightarrow> msgterm" where 
  "extr_ainfo (Mac[macKey asid] (L (Num ts # xs))) = Num ts"
| "extr_ainfo _ = \<epsilon>"

abbreviation ik_auth_ainfo :: "msgterm \<Rightarrow> msgterm" where
  "ik_auth_ainfo \<equiv> id"

text\<open>An authenticated info field is always a number (corresponding to a timestamp).\<close>
abbreviation checkInfo where 
  "checkInfo ainfo \<equiv> (\<exists>ts. ainfo = Num ts)"

text\<open>When observing a hop field, an attacker learns the HVF. UHI is empty and the AHI only contains 
public information that are not terms.\<close>
fun ik_hf :: "SCION_HF \<Rightarrow> msgterm set" where 
  "ik_hf hf = {HVF hf}"

abbreviation no_oracle where "no_oracle \<equiv> (\<lambda> _ _. True)"

text\<open>We now define useful properties of the above definition.\<close>
lemma hf_valid_invert:
  "hf_valid tsn uinfo prev hf mo \<longleftrightarrow>
   ((\<exists>ahi ahi2 ts upif downif asid x upif2 downif2 x2.
     hf = \<lparr>AHI = ahi, UHI = (), HVF = x\<rparr> \<and>
     ASID ahi = asid \<and> ASIF (DownIF ahi) downif \<and> ASIF (UpIF ahi) upif \<and>
     mo = Some \<lparr>AHI = ahi2, UHI = (), HVF = x2\<rparr> \<and>
     ASIF (DownIF ahi2) downif2 \<and> ASIF (UpIF ahi2) upif2 \<and>
     x = Mac[macKey asid] (L [tsn, upif, downif, upif2, downif2, x2]) \<and>
     tsn = Num ts \<and>
     uinfo = \<epsilon>)
 \<or> (\<exists>ahi ts upif downif asid x.
     hf = \<lparr>AHI = ahi, UHI = (), HVF = x\<rparr> \<and>
     ASID ahi = asid \<and> ASIF (DownIF ahi) downif \<and> ASIF (UpIF ahi) upif \<and>
     mo = None \<and>
     x = Mac[macKey asid] (L [tsn, upif, downif]) \<and>
     tsn = Num ts \<and>
     uinfo = \<epsilon>)
    )"
  by(auto elim!: hf_valid.elims)

lemma hf_valid_checkInfo[dest]: "hf_valid ainfo uinfo prev hf z \<Longrightarrow> checkInfo ainfo"
  by(auto simp add: hf_valid_invert)

lemma info_hvf: 
  assumes "hf_valid ainfo uinfo prev m z" "hf_valid ainfo' uinfo' prev' m' z'" "HVF m = HVF m'" 
  shows "ainfo' = ainfo" "m' = m"
  using assms by(auto simp add: hf_valid_invert intro: ahi_eq)

(******************************************************************************)
subsection\<open>Definitions and properties of the added intruder knowledge\<close>
(******************************************************************************)
text\<open>Here we define a @{text "ik_add"} and @{text "ik_oracle"} as being empty, as these features are
not used in this instance model.\<close>

sublocale dataplane_3_directed_defs _ _ _ auth_seg0 hf_valid checkInfo extr extr_ainfo ik_auth_ainfo ik_hf
  by unfold_locales

declare TW.holds_set_list[dest]
declare TW.holds_takeW_is_identity[simp]
declare parts_singleton[dest]

abbreviation ik_add :: "msgterm set" where "ik_add \<equiv> {}"

abbreviation ik_oracle :: "msgterm set" where "ik_oracle \<equiv> {}"

(******************************************************************************)
subsection\<open>Properties of the intruder knowledge, including @{text "ik_add"} and  @{text "ik_oracle"}\<close>
(******************************************************************************)
text\<open>We now instantiate the parametrized model's definition of the intruder knowledge, using the
definitions of @{text "ik_add"} and  @{text "ik_oracle"} from above. We then prove the properties 
that we need to instantiate the @{text "dataplane_3_directed"} locale.\<close>
sublocale
  dataplane_3_directed_ik_defs _ _ _ auth_seg0 hf_valid checkInfo extr extr_ainfo ik_auth_ainfo 
                  ik_hf ik_add ik_oracle no_oracle
  by unfold_locales

lemma ik_auth_hfs_form: "t \<in> parts ik_auth_hfs \<Longrightarrow> \<exists> t' . t = Hash t'"
  apply auto apply(drule parts_singleton)
  by(auto simp add: auth_seg2_def hf_valid_invert dest!: TW.holds_set_list)

declare ik_auth_hfs_def[simp del]

lemma parts_ik_auth_hfs[simp]: "parts ik_auth_hfs = ik_auth_hfs"
  by (auto intro!: parts_Hash ik_auth_hfs_form)

text\<open>This lemma allows us not only to expand the definition of @{term "ik_auth_hfs"}, but also 
to obtain useful properties, such as a term being a Hash, and it being part of a valid hop field.\<close>
lemma ik_auth_hfs_simp: 
  "t \<in> ik_auth_hfs \<longleftrightarrow> (\<exists>t' . t = Hash t') \<and> (\<exists>hf . t = HVF hf
                    \<and> (\<exists>hfs. hf \<in> set hfs \<and> (\<exists>ainfo . (ainfo, hfs) \<in> auth_seg2
                    \<and> (\<exists> prev nxt uinfo. hf_valid ainfo uinfo prev hf nxt))))" (is "?lhs \<longleftrightarrow> ?rhs")
proof 
  assume asm: "?lhs" 
  then obtain ainfo hf hfs where 
    dfs: "hf \<in> set hfs" "(ainfo, hfs) \<in> auth_seg2" "t = HVF hf"
    by(auto simp add: ik_auth_hfs_def)
  then obtain uinfo where "hfs_valid_None ainfo uinfo hfs"  "(ainfo, AHIS hfs) \<in> auth_seg0"
    by(auto simp add: auth_seg2_def)
  then show "?rhs" using asm dfs by(fast intro: ik_auth_hfs_form)
qed(auto simp add: ik_auth_hfs_def)

(******************************************************************************)
subsubsection \<open>Properties of Intruder Knowledge\<close>
(******************************************************************************)
lemma auth_ainfo[dest]: "\<lbrakk>(ainfo, hfs) \<in> auth_seg2\<rbrakk> \<Longrightarrow> \<exists> ts . ainfo = Num ts"
  by(auto simp add: auth_seg2_def)

lemma Num_ik[intro]: "Num ts \<in> ik"
  by(auto simp add: ik_def)
    (auto simp add: auth_seg2_def TW.holds.simps(3) elim!: allE[of _ "[]"]) 

text \<open>There are no ciphertexts (or signatures) in @{term "parts ik"}. Thus, @{term "analz ik"}
and @{term "parts ik"} are identical.\<close>
lemma analz_parts_ik[simp]: "analz ik = parts ik"
  by(rule no_crypt_analz_is_parts)
    (auto simp add: ik_def auth_seg2_def ik_auth_hfs_simp)

lemma parts_ik[simp]: "parts ik = ik"
  by(auto 3 4 simp add: ik_def auth_seg2_def)

lemma key_ik_bad: "Key (macK asid) \<in> ik \<Longrightarrow> asid \<in> bad"
  by(auto simp add: ik_def hf_valid_invert)
    (auto 3 4 simp add: auth_seg2_def ik_auth_hfs_simp hf_valid_invert)

lemma MAC_synth_helper:
  assumes "hf_valid ainfo uinfo prev m z" "HVF m = Mac[Key (macK asid)] j" "HVF m \<in> ik"
  shows "\<exists>hfs. m \<in> set hfs \<and> (ainfo, hfs) \<in> auth_seg2"
proof-
  from assms(2-3) obtain ainfo' uinfo' m' hfs' prev' nxt' where dfs:
"m' \<in> set hfs'" "(ainfo', hfs') \<in> auth_seg2" "hf_valid ainfo' uinfo' prev' m' nxt'" "HVF m = HVF m'"
    by(auto simp add: ik_def ik_auth_hfs_simp)                
  then have "ainfo' = ainfo" "m' = m" using assms(1) by(auto elim!: info_hvf)
  then show ?thesis using dfs assms by auto
qed

text\<open>This definition helps with the limiting the number of cases generated. We don't require it, 
but it is convenient. Given a hop validation field and an asid, return if the hvf has the expected
format.\<close>
definition mac_format :: "msgterm \<Rightarrow> as \<Rightarrow> bool" where 
  "mac_format m asid \<equiv> \<exists> j . m = Mac[macKey asid] j"

text\<open>If a valid hop field is derivable by the attacker, but does not belong to the attacker, then 
the hop field is already contained in the set of authorized segments.\<close>
lemma MAC_synth:
  assumes "hf_valid ainfo uinfo prev m z" "HVF m \<in> synth ik" "mac_format (HVF m) asid"
    "asid \<notin> bad" "checkInfo ainfo"
  shows "\<exists>hfs . m \<in> set hfs \<and> (ainfo, hfs) \<in> auth_seg2"
  using assms
  apply(auto simp add: mac_format_def elim!: MAC_synth_helper dest!: key_ik_bad)
  by(auto simp add: ik_def ik_auth_hfs_simp)

(******************************************************************************)
subsection\<open>Direct proof goals for interpretation of @{text "dataplane_3_directed"}\<close>
(******************************************************************************)

lemma COND_honest_hf_analz:
  assumes "ASID (AHI hf) \<notin> bad" "hf_valid ainfo uinfo prev hf nxt" "ik_hf hf \<subseteq> synth (analz ik)"
    "no_oracle ainfo uinfo"
    shows "ik_hf hf \<subseteq> analz ik"
proof-
  let ?asid = "ASID (AHI hf)"
  from assms(3) have hf_synth_ik: "HVF hf \<in> synth ik" by auto
  from assms(2) have "mac_format (HVF hf) ?asid" "checkInfo ainfo" 
    by(auto simp add: mac_format_def hf_valid_invert)
  then obtain hfs where "hf \<in> set hfs" "(ainfo, hfs) \<in> auth_seg2"
    using assms(1,2,4) hf_synth_ik by(auto dest!: MAC_synth)
  then have "HVF hf \<in> ik"
    using assms(2)
    by(auto simp add: ik_auth_hfs_def intro!: ik_ik_auth_hfs intro!: exI) 
  then show ?thesis by auto
qed

lemma COND_ainfo_analz:
  assumes "hf_valid ainfo uinfo prev hf nxt" and "ik_auth_ainfo ainfo \<in> synth (analz ik)"
  shows "ik_auth_ainfo ainfo \<in> analz ik"
  using assms by(auto simp add: hf_valid_invert)

lemma COND_ik_hf: 
  assumes "hf_valid ainfo uinfo prev hf z" and "HVF hf \<in> ik" and "no_oracle ainfo uinfo"
  shows "\<exists>hfs. hf \<in> set hfs \<and> (ainfo, hfs) \<in> auth_seg2"
proof-
  from assms have "checkInfo ainfo" by auto
  then obtain hfs ainfo where hfs_def: "hf \<in> set hfs" "(ainfo, hfs) \<in> auth_seg2"
  using assms by(auto 3 4 simp add: hf_valid_invert ik_auth_hfs_simp ik_def dest: ahi_eq)
  then obtain hfs ainfo where hfs_def: "hf \<in> set hfs" "(ainfo, hfs) \<in> auth_seg2" by auto
  show ?thesis 
    using hfs_def apply (auto simp add: auth_seg2_def dest!: TW.holds_set_list)
    using hfs_def assms(1) by (auto simp add: auth_seg2_def dest: info_hvf)
  qed

lemma COND_extr_prefix_path:
  "\<lbrakk>hfs_valid ainfo uinfo pre l nxt; nxt = None\<rbrakk> \<Longrightarrow> prefix (extr_from_hd l) (AHIS l)"
  by(induction pre l nxt rule: TW.holds.induct)
    (auto simp add: TW.holds_split_tail TW.holds.simps(1) hf_valid_invert,
     auto split: list.split_asm simp add: hf_valid_invert intro!: ahi_eq elim: ASIF.elims)

lemma COND_path_prefix_extr:
  "prefix (AHIS (hfs_valid_prefix ainfo uinfo pre l nxt))
          (extr_from_hd l)"
  apply(induction pre l nxt rule: TW.takeW.induct[where ?Pa="hf_valid ainfo uinfo"])
  apply(auto simp add: TW.takeW_split_tail TW.takeW.simps(1))
  apply(auto simp add: hf_valid_invert intro!: ahi_eq) (*takes a few seconds*)
    by(auto elim: ASIF.elims)

lemma COND_hf_valid_no_prev:
  "hf_valid ainfo uinfo prev hf z \<longleftrightarrow> hf_valid ainfo uinfo prev' hf z"
  by(auto simp add: hf_valid_invert)

lemma COND_hf_valid_uinfo:
  "\<lbrakk>hf_valid ainfo uinfo pre hf nxt; hf_valid ainfo' uinfo' pre' hf nxt'\<rbrakk> \<Longrightarrow> uinfo' = uinfo"
  by(auto simp add: hf_valid_invert)

(******************************************************************************)
subsection\<open>Instantiation of @{text "dataplane_3_directed"} locale\<close>
(******************************************************************************)
sublocale
  dataplane_3_directed _ _ _ auth_seg0 hf_valid checkInfo extr extr_ainfo ik_auth_ainfo ik_hf ik_add 
            ik_oracle no_oracle
  apply unfold_locales
  using COND_ik_hf COND_honest_hf_analz COND_ainfo_analz COND_extr_prefix_path
  COND_path_prefix_extr COND_hf_valid_no_prev COND_hf_valid_uinfo by auto


end
end
